package com.dzeio.openhealth.connectors

import android.Manifest
import android.app.Activity
import android.content.Intent
import android.content.pm.PackageManager
import android.os.Build
import android.util.Log
import androidx.annotation.RequiresApi
import androidx.core.app.ActivityCompat
import com.dzeio.openhealth.data.weight.Weight
import com.google.android.gms.auth.api.signin.GoogleSignIn
import com.google.android.gms.fitness.Fitness
import com.google.android.gms.fitness.FitnessOptions
import com.google.android.gms.fitness.data.DataPoint
import com.google.android.gms.fitness.data.DataSet
import com.google.android.gms.fitness.data.DataType
import com.google.android.gms.fitness.request.DataReadRequest
import java.text.DateFormat
import java.text.SimpleDateFormat
import java.time.*
import java.util.*
import java.util.concurrent.TimeUnit

class GoogleFit(
    private val activity: Activity,
) : ConnectorInterface {
    companion object {
        const val TAG = "GoogleFitConnector"
    }

    override val sourceID: String = "GoogleFit"

    private val fitnessOptions = FitnessOptions.builder()
        .addDataType(DataType.TYPE_WEIGHT)
        .build()

    private val runningQOrLater =
        Build.VERSION.SDK_INT >= Build.VERSION_CODES.Q

    fun import() {
        checkPermissionsAndRun()
    }

    private fun checkPermissionsAndRun() {
        if (permissionApproved()) {
            signIn(0)
        } else {
            requestRuntimePermissions()
        }
    }

    private fun permissionApproved(): Boolean {
        val approved = if (runningQOrLater) {
            PackageManager.PERMISSION_GRANTED == ActivityCompat.checkSelfPermission(
                activity,
                Manifest.permission.ACCESS_FINE_LOCATION)
        } else {
            true
        }
        return approved
    }

    private fun requestRuntimePermissions() {
        val shouldProvideRationale =
            ActivityCompat.shouldShowRequestPermissionRationale(activity, Manifest.permission.ACCESS_FINE_LOCATION)

        // Provide an additional rationale to the user. This would happen if the user denied the
        // request previously, but didn't check the "Don't ask again" checkbox.
        if (shouldProvideRationale) {
            Log.i(TAG, "Displaying permission rationale to provide additional context.")
//                ProgressDialog.show(activity, "Waiting for authorization...", "")
            ActivityCompat.requestPermissions(activity,
                arrayOf(Manifest.permission.ACCESS_FINE_LOCATION),
                0)
        } else {
            Log.i(TAG, "Requesting permission")
            // Request permission. It's possible this can be auto answered if device policy
            // sets the permission in a given state or the user denied the permission
            // previously and checked "Never ask again".
            ActivityCompat.requestPermissions(activity,
                arrayOf(Manifest.permission.ACCESS_FINE_LOCATION),
                0)
        }
    }

    private fun signIn(requestCode: Int) {
        if (oAuthPermissionsApproved()) {
            Log.d("GoogleFitImporter", "Starting Import")
            internalImportWeight()
        } else {
            Log.d("GoogleFitImporter", "Requesting Permission")
            requestCode.let {
                GoogleSignIn.requestPermissions(
                    activity,
                    it,
                    getGoogleAccount(), fitnessOptions)

            }
        }
    }

    private fun oAuthPermissionsApproved() = GoogleSignIn.hasPermissions(getGoogleAccount(), fitnessOptions)

    /**
     * Gets a Google account for use in creating the Fitness client. This is achieved by either
     * using the last signed-in account, or if necessary, prompting the user to sign in.
     * `getAccountForExtension` is recommended over `getLastSignedInAccount` as the latter can
     * return `null` if there has been no sign in before.
     */
    private fun getGoogleAccount() = GoogleSignIn.getAccountForExtension(activity, fitnessOptions)

    private fun internalImportWeight() {

        val calendar = Calendar.getInstance(TimeZone.getTimeZone("UTC"))
        val now = Date()
        calendar.time = now
        val endTime = calendar.timeInMillis
        calendar.set(Calendar.YEAR, 2013) // Set year to 2013 to be sure to get data from when Google Fit Started to today
        val startTime = calendar.timeInMillis

        val dateFormat = DateFormat.getDateInstance()
        Log.i(TAG, "Range Start: ${dateFormat.format(startTime)}")
        Log.i(TAG, "Range End: ${dateFormat.format(endTime)}")


        runRequest(DataReadRequest.Builder()
            // The data request can specify multiple data types to return, effectively
            // combining multiple data queries into one call.
            // In this example, it's very unlikely that the request is for several hundred
            // datapoints each consisting of a few steps and a timestamp.  The more likely
            // scenario is wanting to see how many steps were walked per day, for 7 days.
//            .aggregate(DataType.AGGREGATE_WEIGHT_SUMMARY)
            .read(DataType.TYPE_WEIGHT)

            // Analogous to a "Group By" in SQL, defines how data should be aggregated.
            // bucketByTime allows for a time span, whereas bucketBySession would allow
            // bucketing by "sessions", which would need to be defined in code.
//            .bucketByTime(1, TimeUnit.MINUTES)
//            .bucketByActivityType(1, TimeUnit.SECONDS)
//            .bucketBySession()
            .setTimeRange(startTime, endTime, TimeUnit.MILLISECONDS)
            .build())

    }

    private fun runRequest(request: DataReadRequest) {
        Fitness.getHistoryClient(activity, GoogleSignIn.getAccountForExtension(activity, fitnessOptions))
            .readData(request)
            .addOnSuccessListener { response ->
                // The aggregate query puts datasets into buckets, so flatten into a
                // single list of datasets
                Log.d(TAG, "Received response! ${response.dataSets.size} ${response.buckets.size} ${response.status}")
                for (dataSet in response.dataSets) {
                    dumpDataSet(dataSet)
                }
//                for (dataSet in response.buckets.flatMap { it.dataSets }) {
//                    dumpDataSet(dataSet)
//                }
            }
            .addOnFailureListener { e ->
                Log.w(TAG,"There was an error reading data from Google Fit", e)
            }
    }

    private fun dumpDataSet(dataSet: DataSet) {
        Log.i(TAG, "Data returned for Data type: ${dataSet.dataType.name} ${dataSet.dataPoints.size}")
        dataSet.dataPoints.forEachIndexed { index, dp ->
            val weight = Weight()
            Log.i(TAG,"Data point:")
            Log.i(TAG,"\tType: ${dp.dataType.name}")
            Log.i(TAG,"\tStart: ${dp.getStartTimeString()}")
            Log.i(TAG,"\tEnd: ${dp.getEndTimeString()}")
            weight.timestamp = dp.getStartTime(TimeUnit.MILLISECONDS)
            weight.source = "GoogleFit"
            for (field in dp.dataType.fields) {
                weight.weight = dp.getValue(field).asFloat()
                Log.i(TAG,"\tField: ${field.name.toString()} Value: ${dp.getValue(field)}")
                callback(weight, (index + 1) == dataSet.dataPoints.size)
            }
        }
    }

    fun DataPoint.getStartTimeString(): String = Date(this.getStartTime(TimeUnit.SECONDS) * 1000L).toLocaleString()

    fun DataPoint.getEndTimeString(): String = Date(this.getEndTime(TimeUnit.SECONDS) * 1000L).toLocaleString()

    override fun onRequestPermissionResult(
        requestCode: Int,
        permission: Array<String>,
        grantResult: IntArray
    ) {
        signIn(requestCode)
    }

    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?) {
        signIn(requestCode)
    }

    private lateinit var callback: (weight: Weight, end: Boolean) -> Unit

    override fun importWeight(callback: (weight: Weight, end: Boolean) -> Unit) {
        this.callback = callback
        checkPermissionsAndRun()
    }

}